# adapted from https://github.com/vwxyzjn/cleanrl
from typing import Tuple

import torch
from entity_gym.env import *
from entity_gym.simple_trace import Tracer

from enn_trainer.agent import PPOAgent

from .config import *


def returns_and_advantages(
    agent: PPOAgent,
    next_obs: VecObs,
    next_done: torch.Tensor,
    rewards: torch.Tensor,
    dones: torch.Tensor,
    values: torch.Tensor,
    gae: float,
    gamma: float,
    gae_lambda: float,
    device: torch.device,
    tracer: Tracer,
) -> Tuple[torch.Tensor, torch.Tensor]:
    # bootstrap value if not done
    next_value = agent.get_auxiliary_head(
        next_obs.features, next_obs.visible, "value", tracer
    ).reshape(1, -1)
    num_steps = values.size(0)
    if gae:
        advantages = torch.zeros_like(rewards).to(device)
        lastgaelam = 0
        for t in reversed(range(num_steps)):
            if t == num_steps - 1:
                nextnonterminal = 1.0 - next_done.float()
                nextvalues = next_value
            else:
                nextnonterminal = 1.0 - dones[t + 1]
                nextvalues = values[t + 1]
            delta = rewards[t] + gamma * nextvalues * nextnonterminal - values[t]
            advantages[t] = lastgaelam = (
                delta + gamma * gae_lambda * nextnonterminal * lastgaelam
            )
        returns = advantages + values
    else:
        returns = torch.zeros_like(rewards).to(device)
        for t in reversed(range(num_steps)):
            if t == num_steps - 1:
                nextnonterminal = 1.0 - next_done
                next_return = next_value
            else:
                nextnonterminal = 1.0 - dones[t + 1]
                next_return = returns[t + 1]
            returns[t] = rewards[t] + gamma * nextnonterminal * next_return
        advantages = returns - values
    # Need to detach here because bug in pytorch that otherwise causes spurious autograd errors and memory leaks when dedicated value function network is used.
    # possibly same cause as this: https://github.com/pytorch/pytorch/issues/71495
    return returns.detach(), advantages.detach()



import json
from typing import Callable, List, Mapping, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
import torch
import torch.distributed as dist
from entity_gym.env import *
from entity_gym.env.add_metrics_wrapper import AddMetricsWrapper
from entity_gym.env.vec_env import Metric
from entity_gym.serialization import SampleRecordingVecEnv
from entity_gym.simple_trace import Tracer
from torch.utils.tensorboard import SummaryWriter

from enn_trainer.agent import PPOAgent
from enn_trainer.config import EnvConfig, EvalConfig, RolloutConfig
from enn_trainer.rollout import Rollout


def run_eval(
    cfg: EvalConfig,
    env_cfg: EnvConfig,
    rollout: RolloutConfig,
    create_env: Callable[[EnvConfig, int, int, int], VecEnv],
    create_opponent: Callable[
        [str, ObsSpace, Mapping[str, ActionSpace], torch.device], PPOAgent
    ],
    agent: PPOAgent,
    device: torch.device,
    tracer: Tracer,
    writer: Optional[SummaryWriter],
    global_step: int,
    rank: int,
    parallelism: int,
) -> None:
    # TODO: metrics are biased towards short episodes
    processes = cfg.processes or rollout.processes
    num_envs = cfg.num_envs or rollout.num_envs

    envs: VecEnv = create_env(
        cfg.env or env_cfg,
        num_envs // parallelism,
        processes,
        rank * num_envs // parallelism,
    )
    obs_space = envs.obs_space()
    action_space = envs.action_space()

    metric_filter: Optional[npt.NDArray[np.bool8]] = None
    if cfg.opponent is not None:
        opponent = create_opponent(cfg.opponent, obs_space, action_space, device)
        if cfg.opponent_only:
            agents: Union[
                PPOAgent, List[Tuple[npt.NDArray[np.int64], PPOAgent]]
            ] = opponent
        else:
            agents = [
                (np.array([2 * i for i in range(num_envs // parallelism // 2)]), agent),
                (
                    np.array([2 * i + 1 for i in range(num_envs // parallelism // 2)]),
                    opponent,
                ),
            ]
            metric_filter = np.arange(num_envs // parallelism) % 2 == 0
    else:
        agents = agent

    envs = AddMetricsWrapper(envs, metric_filter)

    assert num_envs % parallelism == 0, (
        "Number of eval environments must be divisible by parallelism: "
        f"{num_envs} % {parallelism} = {num_envs % parallelism}"
    )

    if cfg.capture_samples:
        envs = SampleRecordingVecEnv(
            envs, cfg.capture_samples, cfg.capture_samples_subsample
        )
    eval_rollout = Rollout(
        envs,
        obs_space=obs_space,
        action_space=action_space,
        agent=agents,
        device=device,
        tracer=tracer,
    )
    _, _, metrics = eval_rollout.run(
        cfg.steps,
        record_samples=False,
        capture_videos=cfg.capture_videos,
        capture_logits=cfg.capture_logits,
    )

    if parallelism > 1:
        serialized_metrics = json.dumps(
            {
                k: {
                    "count": int(v.count),
                    "sum": float(v.sum),
                    "min": float(v.min),
                    "max": float(v.max),
                }
                for k, v in metrics.items()
            }
        )
        metrics_tensor = torch.tensor(
            bytearray(serialized_metrics.encode("utf-8")), dtype=torch.uint8
        )
        metrics = {}
        for metrics_tensor in allgather(metrics_tensor, rank, parallelism):
            for k, v in json.loads(
                metrics_tensor.numpy().tobytes().decode("utf-8")
            ).items():
                m = Metric(**v)
                if k in metrics:
                    metrics[k] += m
                else:
                    metrics[k] = m
    if writer is not None:
        if cfg.capture_videos:
            # save the videos
            writer.add_video(
                f"eval/video",
                torch.tensor(eval_rollout.rendered).permute(1, 0, 4, 2, 3),
                global_step,
                fps=30,
            )

        for name, value in metrics.items():
            writer.add_scalar(f"eval/{name}.mean", value.mean, global_step)
            writer.add_scalar(f"eval/{name}.min", value.min, global_step)
            writer.add_scalar(f"eval/{name}.max", value.max, global_step)
            writer.add_scalar(f"eval/{name}.count", value.count, global_step)
    print(
        f"[eval] global_step={global_step} {'  '.join(f'{name}={value.mean}' for name, value in metrics.items())}"
    )
    envs.close()


def allgather(tensor: torch.Tensor, rank: int, parallelism: int) -> List[torch.Tensor]:
    sizes = [torch.zeros(1, dtype=torch.long) for _ in range(parallelism)]
    dist.all_gather(sizes, torch.tensor([tensor.nelement()], dtype=torch.long))
    alltensors = []
    for src, size in enumerate(sizes):
        if src == rank:
            dist.broadcast(tensor, src)
            alltensors.append(tensor)
        else:
            result = torch.zeros(int(size.item()), dtype=tensor.dtype)
            dist.broadcast(result, src)
            alltensors.append(result)
    return alltensors



import os
from dataclasses import dataclass, field
from typing import Dict, List, Optional

import hyperstate
from hyperstate.schema.rewrite_rule import ChangeDefault, RewriteRule
from rogue_net.rogue_net import RogueNetConfig


@dataclass
class EnvConfig:
    """Environment settings.

    :param kwargs: JSON dictionary with keyword arguments for the environment.
    :param id: The id of the environment.
    :param validate: Perform runtime checks to ensure that the environment correctly implements the interface.
    """

    kwargs: str = "{}"
    id: str = "MoveToOrigin"
    validate: bool = True


@dataclass
class RolloutConfig:
    """Settings for rollout phase of PPO.

    :param steps: The number of steps to run in each environment per policy rollout.
    :param num_envs: The number of parallel game environments.
    :param processes: The number of processes to use to collect env data. The envs are split as equally as possible across the processes.
    """

    steps: int = 16
    num_envs: int = 128
    processes: int = 4


@dataclass
class EvalConfig:
    """Evaluation settings.

    :param interval: Number of environment steps between evaluations.
    :param capture_videos: Render videos of the environments during evaluation.
    :param capture_samples: Write samples from evals to this file.
    :param capture_logits: Record full logits of the agent during evaluation (requires ``capture_samples``).
    :param capture_samples_subsample: Only persist every nth sample, chosen randomly (requires ``capture_samples``).
    :param run_on_first_step: Whether to run an eval on step 0.
    :param env: Settings for the eval environment. If not set, use same settings as rollouts.
    :param num_envs: The number of parallel game environments to use for evaluation. If not set, use same settings as rollouts.
    :param processes: The number of processes used to run the environment. If not set, use same settings as rollouts.
    :param opponent: Path to opponent policy to evaluate against.
    :param opponent_only: Don't evaluate the policy, but instead run the opponent against itself.
    """

    steps: int
    interval: int

    num_envs: Optional[int] = None
    processes: Optional[int] = None
    env: Optional[EnvConfig] = None
    capture_videos: bool = False
    capture_samples: str = ""
    capture_logits: bool = True
    capture_samples_subsample: int = 1
    run_on_first_step: bool = True
    opponent: Optional[str] = None
    opponent_only: bool = False


@dataclass
class PPOConfig:
    """Proximal Policy Optimization settings.

    :param gae: Whether to use generalized advantage estimation for advantage computation.
    :param gamma: Temporal discount factor gamma.
    :param gae_lambda: The lambda for the generalized advantage estimation.
    :param norm_adv: Normalize advantages to 0 mean and 1 std.
    :param clip_coef: The PPO surrogate clipping coefficient.
    :param clip_vloss: Whether to use a clipped loss for the value function.
    :param ent_coef: Coefficient for entropy loss term.
    :param vf_coef: Coefficient for value function loss term.
    :param target_kl: Stop optimization if the KL divergence between the old and new policy exceeds this threshold.
    :param anneal_entropy: Linearly anneal the entropy coefficient from its initial value to 0.
    """

    gae: bool = True
    gamma: float = 0.99
    gae_lambda: float = 0.95
    norm_adv: bool = True
    clip_coef: float = 0.2
    clip_vloss: bool = True
    ent_coef: float = 0.1
    vf_coef: float = 0.5
    target_kl: Optional[float] = None
    anneal_entropy: bool = True


@dataclass
class OptimizerConfig:
    """Optimizer settings.

    :param lr: Adam learning rate.
    :param bs: Batch size.
    :param micro_bs: Micro batch size size used for gradient accumulation. Using a lower micro batch
        size reduces memory usage and performance without affecting training dyanmics.
    :param weight_decay: Adam weight decay.
    :param anneal_lr: Linearly anneal learning rate from initial learning rate to 0.
    :param update_epochs: Number of optimizer passes over each batch of rollout samples.
    :param max_grad_norm: Gradient norm clipping.
    """

    lr: float = 0.001
    lr_warmup_steps: Optional[int] = None
    bs: int = 1024
    weight_decay: float = 0.0
    micro_bs: Optional[int] = None
    anneal_lr: bool = True
    update_epochs: int = 3
    max_grad_norm: float = 2.0


@dataclass
class TrainConfig(hyperstate.Versioned):
    """Training settings.

    :param env: Settings for the environment.
    :param net: Hyperparameters for policy network.
    :param optim: Hyperparameters for optimizer.
    :param ppo: Hyperparameters for PPO.
    :param rollout: Hyperparameters for rollout phase.
    :param eval: Optional evaluation settings.
    :param vf_net: Hyperparameters for value function network (if not set, policy and value function share the same network).
    :param name: The name of the experiment.
    :param seed: Seed of the experiment.
    :param total_timesteps: Total number of timesteps to run for.
    :param max_train_time: Train for at most this many seconds.
    :param torch_deterministic: Sets the value of ``torch.backends.cudnn.deterministic``.
    :param cuda: If ``True``, cuda will be enabled by default.
    :param track: Track experiment metrics with Weights and Biases.
    :param wandb_project_name: Name of the W&B project to log metrics to.
    :param wandb_entity: The entity (team) of the W&B project to log metrics to.
    :param capture_samples: Write all samples collected from environments during training to this file.
    :param capture_logits: Record full logits of the agent (requires ``capture_samples``).
    :param capture_samples_subsample: Only persist every nth sample, chosen randomly (requires ``capture_samples``).
    :param data_dir: Directory to save output from training and logging.
    :param cuda_empty_cache: Empty the torch cuda cache after each optimizer step.
    """

    env: EnvConfig
    net: RogueNetConfig
    optim: OptimizerConfig
    ppo: PPOConfig
    rollout: RolloutConfig
    eval: Optional[EvalConfig] = None
    vf_net: Optional[RogueNetConfig] = None

    name: str = field(default_factory=lambda: os.path.basename(__file__).rstrip(".py"))
    seed: int = 1
    total_timesteps: int = 1000000
    max_train_time: Optional[int] = None
    torch_deterministic: bool = True
    cuda: bool = True
    track: bool = False
    wandb_project_name: str = "enn-ppo"
    wandb_entity: str = "entity-neural-network"
    capture_samples: Optional[str] = None
    capture_logits: bool = False
    capture_samples_subsample: int = 1
    trial: Optional[int] = None
    data_dir: str = "."
    cuda_empty_cache: bool = False

    @classmethod
    def version(clz) -> int:
        return 3

    @classmethod
    def upgrade_rules(clz) -> Dict[int, List[RewriteRule]]:
        return {
            0: [
                ChangeDefault(
                    field=("net", "relpos_encoding", "per_entity_values"),
                    old_default=True,
                    new_default=False,
                ),
                ChangeDefault(
                    field=("vf_net", "relpos_encoding", "per_entity_values"),
                    old_default=True,
                    new_default=False,
                ),
            ],
            1: [
                ChangeDefault(field=("net", "n_layer"), old_default=1, new_default=2),
                ChangeDefault(field=("net", "d_model"), old_default=64, new_default=32),
                ChangeDefault(
                    field=("optim", "lr"), old_default=0.00025, new_default=0.001
                ),
                ChangeDefault(field=("optim", "bs"), old_default=128, new_default=1024),
                ChangeDefault(
                    field=("ppo", "ent_coef"), old_default=0.01, new_default=0.1
                ),
                ChangeDefault(
                    field=("ppo", "anneal_entropy"), old_default=False, new_default=True
                ),
                ChangeDefault(
                    field=("rollout", "steps"), old_default=128, new_default=16
                ),
                ChangeDefault(
                    field=("rollout", "num_envs"), old_default=4, new_default=128
                ),
                ChangeDefault(
                    field=("rollout", "processes"), old_default=1, new_default=4
                ),
                ChangeDefault(
                    field=("vf_net", "n_layer"), old_default=1, new_default=2
                ),
                ChangeDefault(
                    field=("vf_net", "d_model"), old_default=64, new_default=32
                ),
            ],
            2: [
                ChangeDefault(
                    field=("total_timesteps",), old_default=25000, new_default=1000000
                ),
            ],
        }


if __name__ == "__main__":
    hyperstate.schema_evolution_cli(TrainConfig)



from os import listdir
from pathlib import Path

import hyperstate
from hyperstate.schema.schema_change import Severity
from hyperstate.schema.schema_checker import SchemaChecker
from hyperstate.schema.types import load_schema

from enn_trainer.config import TrainConfig


def test_schema() -> None:
    schema_files = Path(__file__).parent.parent.parent / "config-schema.ron"
    path = str(schema_files)
    print(path)
    with open(path) as f:
        f.read()
    old = load_schema(path)
    checker = SchemaChecker(old, TrainConfig)
    if checker.severity() >= Severity.WARN:
        checker.print_report()
    assert checker.severity() == Severity.INFO


def test_configs() -> None:
    config_dir = Path(__file__).parent.parent.parent / "configs" / "entity-gym"
    for config_file in listdir(config_dir):
        print(config_file)
        hyperstate.load(TrainConfig, config_dir / config_file)



from typing import Dict, List, Mapping, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
import ragged_buffer
import torch
from entity_gym.env import *
from entity_gym.env.vec_env import Metric
from entity_gym.ragged_dict import RaggedActionDict, RaggedBatchDict
from entity_gym.serialization.sample_recorder import SampleRecordingVecEnv
from entity_gym.simple_trace import Tracer
from ragged_buffer import RaggedBufferBool, RaggedBufferF32, RaggedBufferI64
from rogue_net.rogue_net import tensor_dict_to_ragged

from enn_trainer.agent import PPOAgent


class Rollout:
    def __init__(
        self,
        envs: VecEnv,
        obs_space: ObsSpace,
        action_space: Mapping[str, ActionSpace],
        agent: Union[PPOAgent, List[Tuple[npt.NDArray[np.int64], PPOAgent]]],
        device: torch.device,
        tracer: Tracer,
        value_function: Optional[PPOAgent] = None,
    ) -> None:
        self.envs = envs
        self.obs_space = obs_space
        self.action_space = action_space
        self.device = device
        self.agent = agent
        self.value_function = value_function
        self.tracer = tracer

        self.global_step = 0
        self.next_obs: Optional[VecObs] = None
        self.next_done: Optional[torch.Tensor] = None
        self.rewards = torch.zeros(0)
        self.dones = torch.zeros(0)
        self.values = torch.zeros(0)
        self.entities: RaggedBatchDict[np.float32] = RaggedBatchDict(RaggedBufferF32)
        self.visible: RaggedBatchDict[np.bool_] = RaggedBatchDict(RaggedBufferBool)
        self.action_masks = RaggedActionDict()
        self.actions = RaggedBatchDict(RaggedBufferI64)
        self.logprobs = RaggedBatchDict(RaggedBufferF32)

        self.rendered_frames: List[npt.NDArray[np.uint8]] = []
        self.rendered: Optional[npt.NDArray[np.uint8]] = None

    def run(
        self,
        steps: int,
        record_samples: bool,
        capture_videos: bool = False,
        capture_logits: bool = False,
    ) -> Tuple[VecObs, torch.Tensor, Dict[str, Metric]]:
        """
        Run the agent for a number of steps. Returns next_obs, next_done, and a dictionary of statistics.
        """
        if record_samples:
            if self.rewards.shape != (steps, len(self.envs)):
                self.rewards = torch.zeros((steps, len(self.envs)), device=self.device)
                self.dones = torch.zeros((steps, len(self.envs)), device=self.device)
                self.values = torch.zeros((steps, len(self.envs)), device=self.device)
            self.entities.clear()
            self.visible.clear()
            self.action_masks.clear()
            self.actions.clear()
            self.logprobs.clear()
        if isinstance(self.agent, list):
            allindices = np.concatenate([indices for indices, _ in self.agent])
            invindex = np.zeros_like(allindices, dtype=np.int64)
            for i, index in enumerate(allindices):
                invindex[index] = i
        else:
            invindex = np.array([], dtype=np.int64)

        metrics: Dict[str, Metric] = {}

        if self.next_obs is None or self.next_done is None:
            next_obs = self.envs.reset(self.obs_space)
            next_done = torch.zeros(len(self.envs), device=self.device)
        else:
            next_obs = self.next_obs
            next_done = self.next_done

        if capture_videos:
            self.rendered_frames.append(self.envs.render(mode="rgb_array"))

        for step in range(steps):
            self.global_step += len(self.envs)

            if record_samples:
                # TODO: this breaks if entity missing on some steps, need merge full VecObs
                self.entities.extend(next_obs.features)
                self.visible.extend(next_obs.visible)
                self.action_masks.extend(next_obs.action_masks)
                self.dones[step] = next_done

            with torch.no_grad(), self.tracer.span("forward"):
                if isinstance(self.agent, list):
                    actions = []
                    for env_indices, agent in self.agent:
                        a = agent.get_action_and_auxiliary(
                            {
                                name: feats[env_indices]
                                for name, feats in next_obs.features.items()
                            },
                            {
                                name: visible[env_indices]
                                for name, visible in next_obs.visible.items()
                            },
                            {
                                name: mask[env_indices]
                                for name, mask in next_obs.action_masks.items()
                            },
                            self.tracer,
                        )[0]
                        actions.append((env_indices, a))
                    action = {}
                    for name in self.action_space.keys():
                        action[name] = ragged_buffer.cat([a[1][name] for a in actions])[
                            invindex
                        ]
                else:
                    (
                        action,
                        probs_tensor,
                        _,
                        actor_counts,
                        aux,
                        logits,
                    ) = self.agent.get_action_and_auxiliary(
                        next_obs.features,
                        next_obs.visible,
                        next_obs.action_masks,
                        tracer=self.tracer,
                    )
                    logprob = tensor_dict_to_ragged(
                        RaggedBufferF32, probs_tensor, actor_counts
                    )
            if record_samples:
                if self.value_function is None:
                    value = aux["value"]
                else:
                    # TODO: can ignore `visible` here, allow for full attention across all entities
                    value = self.value_function.get_auxiliary_head(
                        next_obs.features, next_obs.visible, "value", tracer=self.tracer
                    )

                # Need to detach here because bug in pytorch that otherwise causes spurious autograd errors and memory leaks when dedicated value function network is used.
                # possibly same cause as this: https://github.com/pytorch/pytorch/issues/71495
                self.values[step] = value.detach().flatten()
                self.actions.extend(action)
                self.logprobs.extend(logprob)

            if capture_videos:
                self.rendered_frames.append(self.envs.render(mode="rgb_array"))

            with self.tracer.span("step"):
                if isinstance(self.envs, SampleRecordingVecEnv):
                    if capture_logits:
                        ragged_logits: Optional[
                            Dict[str, RaggedBufferF32]
                        ] = tensor_dict_to_ragged(
                            RaggedBufferF32,
                            {k: v.squeeze(1) for k, v in logits.items()},
                            actor_counts,
                        )
                    else:
                        ragged_logits = None
                    next_obs = self.envs.act(
                        action, self.obs_space, logprob, ragged_logits
                    )
                else:
                    next_obs = self.envs.act(action, self.obs_space)

            if record_samples:
                with self.tracer.span("reward_done_to_device"):
                    self.rewards[step] = (
                        torch.tensor(next_obs.reward).to(self.device).view(-1)
                    )
                    next_done = torch.tensor(next_obs.done).to(self.device).view(-1)

            for mname, mvalue in next_obs.metrics.items():
                if mname in metrics:
                    metrics[mname] += mvalue
                else:
                    metrics[mname] = mvalue

        self.next_obs = next_obs
        self.next_done = next_done

        if capture_videos:
            self.rendered = np.stack(self.rendered_frames)

        return next_obs, next_done, metrics



from typing import Dict, Iterator, Mapping, Optional, Protocol, Tuple

import entity_gym.runner
import numpy as np
import numpy.typing as npt
import torch
from entity_gym.env import Action, Observation, VecActionMask
from entity_gym.env.env_list import action_index_to_actions
from entity_gym.env.environment import ActionName
from entity_gym.env.vec_env import batch_obs
from entity_gym.simple_trace import Tracer
from ragged_buffer import RaggedBufferBool, RaggedBufferF32, RaggedBufferI64
from rogue_net.rogue_net import RogueNet


class PPOAgent(Protocol):
    def get_action_and_auxiliary(
        self,
        entities: Mapping[str, RaggedBufferF32],
        visible: Mapping[str, RaggedBufferBool],
        action_masks: Mapping[str, VecActionMask],
        tracer: Tracer,
        prev_actions: Optional[Dict[str, RaggedBufferI64]] = None,
    ) -> Tuple[
        Dict[str, RaggedBufferI64],  # actions
        Dict[str, torch.Tensor],  # action probabilities
        Dict[str, torch.Tensor],  # entropy
        Dict[str, npt.NDArray[np.int64]],  # number of actors in each frame
        Dict[str, torch.Tensor],  # auxiliary head values
        Dict[str, torch.Tensor],  # full logits
    ]:
        ...

    def to(self, device: torch.device) -> "PPOAgent":
        ...

    def parameters(self, recurse: bool = True) -> Iterator[torch.nn.Parameter]:
        ...

    def get_auxiliary_head(
        self,
        entities: Mapping[str, RaggedBufferF32],
        visible: Mapping[str, RaggedBufferBool],
        head_name: str,
        tracer: Tracer,
    ) -> torch.Tensor:
        ...


class RogueNetAgent(entity_gym.runner.Agent):
    """
    Wraps a rogue_net network, exposing an entity_gym Agent interface.

    :param agent: The underlying RogueNet.
    """

    def __init__(self, agent: RogueNet):
        self.agent = agent

    def act(self, obs: Observation) -> Tuple[Dict[ActionName, Action], float]:
        vec_obs = batch_obs([obs], self.agent.obs_space, self.agent.action_space)
        with torch.no_grad():
            act_indices, _, _, _, aux, logits = self.agent.get_action_and_auxiliary(
                vec_obs.features,
                vec_obs.visible,
                vec_obs.action_masks,
                tracer=Tracer(False),
            )
        actions = action_index_to_actions(
            self.agent.obs_space,
            self.agent.action_space,
            act_indices,
            obs,
            probs={k: l.exp().cpu().numpy() for k, l in logits.items()},
        )
        return actions, float(aux["value"].item())



from typing import Mapping, Tuple

import torch
from entity_gym.env.environment import ActionName
from entity_gym.simple_trace import Tracer
from ragged_buffer import RaggedBufferF32

from enn_trainer.config import PPOConfig


def ppo_loss(
    cfg: PPOConfig,
    newlogprob: Mapping[ActionName, torch.Tensor],
    oldlogprob: Mapping[ActionName, RaggedBufferF32],
    advantages: torch.Tensor,
    device: torch.device,
    tracer: Tracer,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    with tracer.span("ratio"):
        logratio = {
            k: newlogprob[k]
            - torch.tensor(oldlogprob[k].as_array(), device=device).squeeze(-1)
            for k in newlogprob.keys()
        }
        ratio = {k: l.exp() for k, l in logratio.items()}

    with torch.no_grad(), tracer.span("kl"):
        # calculate approx_kl http://joschu.net/blog/kl-approx.html
        # old_approx_kl = (-logratio).mean()
        # TODO: mean across everything rather than nested mean? or do summation over different actions?
        approx_kl = torch.tensor(
            [
                ((_ratio - 1) - _logratio).mean()
                for (_ratio, _logratio) in zip(ratio.values(), logratio.values())
            ]
        ).mean()
        clipfrac = torch.tensor(
            [
                ((_ratio - 1.0).abs() > cfg.clip_coef).float().mean().item()
                for _ratio in ratio.values()
            ]
        ).mean()

    # TODO: not invariant to microbatch size, should be normalizing full batch or minibatch instead
    if cfg.norm_adv:
        assert len(advantages) > 1, "Can't normalize advantages with minibatch size 1"
        advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-8)

    # TODO: we can elide the index op and get better performance when there is exactly one actor per action
    # TODO: we can reuse the mb_advantages across all actions that have the same number of actors
    # TODO: what's the correct way of combining loss from multiple actions/actors on the same timestep? should we split the advantages across actions/actors?
    with tracer.span("broadcast_advantages"):
        # Broadcast the advantage value from each timestep to all actors/actions on that timestep
        bc_mb_advantages = {
            action_name: advantages[
                torch.tensor(
                    _b_logprobs.indices(dim=0).as_array().flatten(), device=device
                )
            ]
            for action_name, _b_logprobs in oldlogprob.items()
        }

    # Policy loss
    with tracer.span("policy_loss"):
        pg_loss1 = torch.cat(
            [
                -_advantages * _ratio.flatten()
                for _advantages, _ratio in zip(
                    bc_mb_advantages.values(), ratio.values()
                )
            ]
        )
        pg_loss2 = torch.cat(
            [
                -_advantages
                * torch.clamp(
                    _ratio.flatten(),
                    1 - cfg.clip_coef,
                    1 + cfg.clip_coef,
                )
                for _advantages, _ratio in zip(
                    bc_mb_advantages.values(), ratio.values()
                )
            ]
        )
        pg_loss = torch.max(pg_loss1, pg_loss2).mean()

    return pg_loss, clipfrac, approx_kl


def value_loss(
    cfg: PPOConfig,
    newvalue: torch.Tensor,
    returns: torch.Tensor,
    oldvalue: torch.Tensor,
    tracer: Tracer,
) -> torch.Tensor:
    with tracer.span("value_loss"):
        newvalue = newvalue.view(-1)
        if cfg.clip_vloss:
            v_loss_unclipped = (newvalue - returns) ** 2
            v_clipped = oldvalue + torch.clamp(
                newvalue - oldvalue,
                -cfg.clip_coef,
                cfg.clip_coef,
            )
            v_loss_clipped = (v_clipped - returns) ** 2
            v_loss_max = torch.max(v_loss_unclipped, v_loss_clipped)
            v_loss = 0.5 * v_loss_max.mean()
        else:
            v_loss = 0.5 * ((newvalue - returns) ** 2).mean()
    return v_loss  # type: ignore
